from click.testing import CliRunner

from rio.commands import cmd_begin
from rio.commands.cmd_end import cli
from rio.utilities import errors


def test_end_nonlocal():
    """
    Tests that not passing the local flag tells one to contact ChainOpt support.
    """
    runner = CliRunner()
    result = runner.invoke(cli)
    assert isinstance(result.exception, errors.NoLocalFlagError)


# TODO: Not working due to unintended behaviour of RIO end
# def test_end_not_running():
#     """
#     Tests to ensure that when there's no image running locally that there will be the proper output.
#     """
#     runner = CliRunner()
#     # First runs the command to ensure that if RIO is running we stop it first.
#     runner.invoke(cli, "-l")
#     result = runner.invoke(cli, "-l")
#     assert not result.exception
#     assert "RIO was not running. To restart, simply run the 'rio begin' command.\n" in result.output


def test_end_base():
    """
    Base case where we end the local rio instance that was already running.
    """
    runner = CliRunner()
    # First ensure that RIO is running
    runner.invoke(cmd_begin.cli, "-l", input="y\n")
    result = runner.invoke(cli, "-l", input="y\n")
    assert not result.exception
    assert "RIO has ended. To restart, simply run the 'rio begin -l' command.\n" in result.output



from click.testing import CliRunner

from rio.commands import cmd_deploy, cmd_undeploy
from rio.commands.cmd_logs import cli
from rio.utilities import errors


def test_logs_nonlocal():
    """
    Tests that not passing the local flag tells one to contact ChainOpt support.
    """
    runner = CliRunner()
    result = runner.invoke(cli, "afsdl;kj438094fglk")
    assert isinstance(result.exception, errors.NoLocalFlagError)


def test_logs_no_package():
    """
    Tests what happens when no package name is passed through
    """
    runner = CliRunner()
    result = runner.invoke(cli)
    assert isinstance(result.exception, SystemExit)
    assert result.exit_code == 2


def test_logs_missing_package():
    """
    Tests when a package name is given that doesn't exist
    """
    runner = CliRunner()
    # Undeploys then redeploys the sample project to ensure that something is running
    runner.invoke(cmd_undeploy.cli, ["-l", "-p", "myProject"], input="y\n")
    runner.invoke(cmd_deploy.cli, ["-l", r"..\samples\myProject"])
    result = runner.invoke(cli, ["-l", "rio-api"])
    assert isinstance(result.exception, errors.PackageExistenceError)



from click.testing import CliRunner
import os
from rio.commands.cmd_begin import cli
from rio.utilities import errors


def test_cmd_begin():
    """
    Tests that simply running with the local flag starts RIO
    """
    runner = CliRunner()
    result = runner.invoke(cli, ["-l"], input="y\n")
    assert not result.exception
    assert "Running RIO" in result.output
    assert "Loading credentials from: " in result.output


def test_cmd_begin_nonlocal():
    """
    Tests that not passing the local flag tells one to contact ChainOpt support.
    """
    runner = CliRunner()
    result = runner.invoke(cli)
    assert isinstance(result.exception, errors.NoLocalFlagError)


def test_begin_good_creds():
    """
    Tests if passing through good creds successfully deploys RIO fully.
    """
    runner = CliRunner()
    # We will pass through  previously saved credentials in the Documents folder, set up as stated in the README
    cred_path = os.path.join(os.path.join(os.path.expanduser("~"), "Documents"), "credentials.yaml")
    result = runner.invoke(cli, ["-l", "-f", cred_path])

    # Checks for    1) no errors
    #               2) completion of beginning the API container
    #               3) Loading the credentials from the file
    #               4) Ensuring it didn't try to load credentials from previously saved location
    assert not result.exception
    assert "Running RIO v" in result.output
    assert "Loading credentials from file:" in result.output
    assert "These credentials are used to check" not in result.output


def test_begin_bad_creds():
    """
    Tests handling of bad given Docker credentials
    """
    runner = CliRunner()
    result = runner.invoke(cli, ["-l", "-f", r"samples\credentials.yaml"])
    # Checks for    1) no errors
    #               2) Failed creds
    #               3) Loading the credentials from the default saved file
    #               4) Ensure Credentials succeed
    #               5) Ensure API begins
    assert not result.exception
    assert "Provided credentials file failed to authenticate." in result.output
    assert "Loading credentials from: " in result.output
    assert "Saved credentials file failed to authenticate." not in result.output
    assert "Running RIO v" in result.output


def test_begin_bad_creds_input_no_saved():
    """
    Tests what happens when all credentials served are bad
    """
    runner = CliRunner()
    # We're going to need to mess with the saved file, then change it back at the end so that we can emulate not having
    # the previously saved credentials file
    cred_path = os.path.join(os.path.join(os.path.expanduser("~"), "Documents"), "credentials.yaml")
    new_path = cred_path + ".old"
    os.rename(cred_path, new_path)
    # Intentionally serve up a bad username, password, and email
    result = runner.invoke(cli, ["-l", "-f", r"samples\credentials.yaml"], input="username\npassword\nemail\n")
    os.rename(new_path, cred_path)
    # Checks for    1) Correct exception raised
    #               2) Provided credentials don't authenticate
    #               3) RIO does not run
    assert isinstance(result.exception, errors.BadCredError)
    assert "Provided credentials file failed to authenticate." in result.output
    assert "Running RIO" not in result.output



from click.testing import CliRunner

from rio.commands import cmd_undeploy, cmd_deploy
from rio.commands.cmd_start import cli
from rio.utilities import errors


def test_start_nonlocal():
    """
    Tests that not passing the local flag tells one to contact ChainOpt support.
    """
    runner = CliRunner()
    result = runner.invoke(cli)
    assert isinstance(result.exception, errors.NoLocalFlagError)


def test_start_bad_package():
    """
    Tests when a bad package name is passed as input
    """
    runner = CliRunner()
    # Undeploys then redeploys the sample project to ensure that something is running
    runner.invoke(cmd_undeploy.cli, ["-l", "-p", "myProject"], input="y\n")
    runner.invoke(cmd_deploy.cli, ["-l", r"..\samples\myProject"])
    result = runner.invoke(cli, ["-l", "-p", "rio-api"])
    assert isinstance(result.exception, errors.PackageExistenceError)


def test_start_no_running_packages():
    """
    Tests when there are no packages running.
    """
    runner = CliRunner()
    # Undeploys ALL packages
    runner.invoke(cmd_undeploy.cli, ["-l", "--all"], input="y\n")
    result = runner.invoke(cli, ["-l", "-p", "rio-api"])
    assert isinstance(result.exception, errors.NoRunningPackagesError)



class BadCredError(SystemExit):
    """
    Handles passing through bad credentials to Docker
    """

    def __init__(self, message="All credentials given failed to authenticate."):
        super().__init__(message)


class DockerNotRunningError(SystemExit):
    """
    Handles trying to run RIO locally without Docker running
    """

    def __init__(self, message="Docker cannot be found running anywhere on this machine."):
        super().__init__(message)


class DockerHubAccountError(SystemExit):
    """
    Handles when the user is not logged in with the correct Docker Hub account.
    """

    def __init__(self, message="The Docker Hub account used does not have permission to the rio-api image. "
                               "Please try another accouunt or try again."):
        super().__init__(message)


class NoLocalFlagError(SystemExit):
    """
    Handles when local flag is not set

    ***SET TO BE DEPRECATED UPON RELEASING CLOUD SUPPORT THROUGH CLI***
    """

    def __init__(self, message="Please email contact@chainopt.com for remote deployment services."):
        super().__init__(message)


class LoggingError(SystemExit):
    """
    Handles when creation of logs has failed.
    """

    def __init__(self, message="Logs could not be generated. Please check your code and retry or contact support."):
        super().__init__(message)


class PackageDeletionError(SystemExit):
    """
    Handles when deleting a package has failed.
    """

    def __init__(self, message="Unable to delete package."):
        super().__init__(message)


class DuplicatePackageError(SystemExit):
    """
    Handles when a user attempts to deploy a package with a duplicate name
    and chooses not to overwrite the package with the new deployment.
    """

    def __init__(self, message="A deployed package already exists with that name. "
                               "Please attempt again with a new package name."):
        super().__init__(message)


class BadPortError(SystemExit):
    """
    Handles when a user inputs a port that cannot be used.
    """

    def __init__(self, message="Please try again and enter a port number between 1024 and 65535. You also may not "
                               "specify a port and it will be assigned automatically."):
        super().__init__(message)


class UploadAPIError(SystemExit):
    """
    Handles when the RIO upload API is non-responsive.
    """

    def __init__(self, message="Upload API unavailable!"):
        super().__init__(message)


class DeployAPIError(SystemExit):
    """
    Handles when the RIO deploy API is non-responsive.
    """

    def __init__(self, message="Deploy API unavailable!"):
        super().__init__(message)


class APIError(SystemExit):
    # TODO: Combine Upload and Deploy API errors into this singular error
    """
    Handles issues with the RIO-API backend.
    """

    def __init__(self, module="", message="API unavailable!"):
        if module:
            super().__init__(module + " " + message)
        else:
            super().__init__(message)


class PackagePathError(SystemExit):
    """
    Handles bad OS paths being input for deployments.
    """

    def __init__(self, message="Unable to find package in location given. Please check the path and try again."):
        super().__init__(message)


class PackageExistenceError(SystemExit):
    """
    Handles when a package name is given that does not exist in Docker/RIO-API
    """

    def __init__(self, package_name, message=" not found. Please check the package name and retry."):
        super().__init__(package_name + message)


class NoRunningPackagesError(SystemExit):
    """
    Handles when a request has been made but no packages are running.
    """

    def __init__(self, message="No running packages."):
        super().__init__(message)


class MissingPackageInputError(SystemExit):
    """
    Handles when a function that requires a package or --all gets neither input
    """

    def __init__(self, message="Please specify a package name or --all."):
        super().__init__(message)


class PackagingError(SystemExit):
    """
    Handles when creating a zipped version of the package fails
    """

    def __init__(self, message="Failed to create .zip version of package. Please check package name."):
        super().__init__(message)


class PackageStopError(SystemExit):
    """
    Handles when the API fails to stop a package
    """

    def __init__(self, package_name):
        message = f"Unable to stop {package_name} package. Please retry or report this as a bug."
        super().__init__()



from click.testing import CliRunner

from rio.commands.cmd_list import cli
from rio.utilities import errors


def test_list_nonlocal():
    """
    Tests that not passing the local flag tells one to contact ChainOpt support.
    """
    runner = CliRunner()
    result = runner.invoke(cli)
    assert isinstance(result.exception, errors.NoLocalFlagError)



# Copyright 2021 Chainopt LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import socket
import time
from contextlib import closing
from pathlib import Path

import click
import docker
import numpy as np
import pandas as pd
import requests
import stdiomask
import yaml
from docker import APIClient
from docker.utils import kwargs_from_env

from rio.utilities import loader
from rio.utilities.errors import *

pkg_lib_path = os.path.join(os.path.join(os.path.expanduser("~"), "Documents"), "pkg_lib.pkl")


def load_yaml(stream) -> dict:
    try:
        cred_yaml = yaml.load(stream, Loader=yaml.Loader)
    except yaml.YAMLError as exception:
        cred_yaml = None
    return cred_yaml


def delete_both(base_url, package_name):
    delete_container(base_url, package_name)
    delete_package(base_url, package_name)
    return


def delete_package(base_url, package_name):
    response_delete = requests.get(base_url + '/packages/delete/?package_name=' + package_name)

    if response_delete.status_code == 200:
        pass
    else:
        click.echo("Unable to delete package: " + package_name)
    return


def delete_container(base_url, package_name):
    response_remove_container = requests.get(base_url + '/containers/remove/?package_name=' + package_name)

    if response_remove_container.status_code == 200:
        pass
    else:
        click.echo("Unable to shut down API for package: " + package_name)

    return


def check_port(port):
    if port:
        if 1023 < int(port) < 65535:
            if not is_port_in_use(int(port)):
                port_is_good = True
            else:
                click.echo("The port " + port + " is already in use.")
                port_is_good = False
        else:
            port_is_good = False
    else:
        port = find_free_port()
        port_is_good = True
    return port, port_is_good


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('localhost', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def is_port_in_use(port) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


def get_rio_url() -> str:
    try:
        client = APIClient(**kwargs_from_env())
    except docker.errors.DockerException:
        click.echo("Docker is not running!")
        raise
    rio_port = None
    for i in client.containers():
        if i['Names'][0] == '/rio-api':
            rio_port = i['Ports'][0]['PublicPort']
    if rio_port:
        base_url = 'http://localhost:' + str(rio_port)
    else:
        base_url = None
    return base_url


def get_package_url(package_name: str):
    response_running_containers = requests.get(get_rio_url() + '/containers/?include_stopped=true')
    dfAllContainers = pd.read_json(response_running_containers.text, orient='records')
    dfRunningContainers = dfAllContainers[dfAllContainers['State'] == 'running']
    launch_url = 'http://localhost:' + str(dfRunningContainers[dfRunningContainers['Names'] == package_name]
                                           .iloc[0]['Ports'].split(":")[1].split("-")[0])
    return launch_url


def get_package_state(package_name: str):
    response_all_containers = requests.get(get_rio_url() + '/containers/?include_stopped=true')
    dfAllContainers = pd.read_json(response_all_containers.text, orient='records')
    package_state = ""
    if package_name in dfAllContainers.Names.values:
        package_state = str(dfAllContainers[dfAllContainers['Names'] == package_name].iloc[0]['State'])
    return package_state


def tail(f, lines=20):
    total_lines_wanted = lines

    BLOCK_SIZE = 1024
    f.seek(0, 2)
    block_end_byte = f.tell()
    lines_to_go = total_lines_wanted
    block_number = -1
    blocks = []
    while lines_to_go > 0 and block_end_byte > 0:
        if (block_end_byte - BLOCK_SIZE > 0):
            f.seek(block_number * BLOCK_SIZE, 2)
            blocks.append(f.read(BLOCK_SIZE))
        else:
            f.seek(0, 0)
            blocks.append(f.read(block_end_byte))
        lines_found = blocks[-1].count(b'\n')
        lines_to_go -= lines_found
        block_end_byte -= BLOCK_SIZE
        block_number -= 1
    all_read_text = b''.join(reversed(blocks))
    return b'\n'.join(all_read_text.splitlines()[-total_lines_wanted:])


def uri_exists_stream(uri: str) -> bool:
    try:
        with requests.get(uri, stream=True) as response:
            try:
                response.raise_for_status()
                return True
            except requests.exceptions.HTTPError:
                return False
    except requests.exceptions.ConnectionError:
        return False


def begin(local, file):
    if local:
        try:
            client, loading, docker_creds = login(file)
        except BadCredError:
            raise
        except docker.errors.DockerException:
            raise DockerNotRunningError

        # specify version here or use 'latest'
        version = '1.0.2'

        client.pull('chainopt/rio-api', tag=version)

        # Custom function in utils to find a free port on local
        rio_port = find_free_port()

        # check if RIO is already running locally. If it is, then remove it.
        myContainers = client.containers(all=True)
        for i in myContainers:
            if i['Names'][0] == '/rio-api':
                client.remove_container(i, force=True)
                click.echo("\n ...Restarting RIO")

        # create the rio-api local deployment docker container and specify ports and volume bindings.
        try:
            container = client.create_container(name='rio-api', image='chainopt/rio-api:' + version, stdin_open=True,
                                                tty=True,
                                                environment=docker_creds,
                                                volumes=['/var/run/docker.sock', '/var/rio-packages'], ports=[80],
                                                host_config=client.create_host_config(binds={
                                                    '/var/run/docker.sock': {'bind': '/var/run/docker.sock',
                                                                             'mode': 'rw', },
                                                    '/var/rio-packages': {'bind': '/var/rio-packages', 'mode': 'rw', }},
                                                    port_bindings={80: rio_port}))
        except docker.errors.ImageNotFound:
            raise DockerHubAccountError

        client.start(container)

        while not uri_exists_stream(get_rio_url()):
            # wait for rio to start up
            pass

        loading.stop()

        click.echo("Running RIO v" + version)

    else:
        raise NoLocalFlagError()

    return


def login(file: str):
    """
    Handles the sequence of trying different credential files to gain access to DockerHub
    """
    # Get local docker client setup
    try:
        client = APIClient(**kwargs_from_env())
    except docker.errors.DockerException:
        raise
    # Check if a filename was passed through
    if file:
        click.echo("Loading credentials from file: " + file)
        docker_creds = load_yaml(open(file, "r", encoding="utf-8"))
        try:
            client.login(username=docker_creds.get("docker_user"), password=docker_creds.get("docker_pw"),
                         email=docker_creds.get("docker_email"), registry='https://index.docker.io/v1/')
            load_object = loader.Loader("Setting up local deployment services..",
                                        "RIO is now ready to deploy your model.",
                                        0.05).start()
            return client, load_object, docker_creds
        except requests.exceptions.HTTPError:
            click.echo("Provided credentials file failed to authenticate.")

    # Check if credentials were previously successfully provided.
    cred_path = os.path.join(os.path.join(os.path.expanduser("~"), "Documents"), "credentials.yaml")
    if os.path.exists(cred_path):
        click.echo("Loading credentials from: " + cred_path)
        docker_creds = load_yaml(open(cred_path, "r", encoding="utf-8"))
        try:
            client.login(username=docker_creds.get("docker_user"), password=docker_creds.get("docker_pw"),
                         email=docker_creds.get("docker_email"), registry='https://index.docker.io/v1/')
            load_object = loader.Loader("Setting up local deployment services..",
                                        "RIO is now ready to deploy your model.",
                                        0.05).start()
            return client, load_object, docker_creds
        except requests.exceptions.HTTPError:
            click.echo("Saved credentials failed to authenticate.")
    # Final solution is to ask for credentials
    click.echo("These credentials are used to check if you have access to ChainOpt artifacts on Docker Hub.")
    docker_creds = dict()
    docker_creds["docker_user"] = input("Enter your Docker Username: ")
    docker_creds["docker_pw"] = stdiomask.getpass(prompt='Enter your Docker Password: ', mask='*')
    docker_creds["docker_email"] = input("Enter your Docker email: ")
    try:
        client.login(username=docker_creds.get("docker_user"), password=docker_creds.get("docker_pw"),
                     email=docker_creds.get("docker_email"), registry='https://index.docker.io/v1/')
        # Only save credentials to a file if they work.
        with open(cred_path, 'w') as f:
            yaml.dump(docker_creds, f)
            click.echo("Your credentials have been stored at: " + cred_path)
            f.close()
        load_object = loader.Loader("Setting up local deployment services..", "RIO is now ready to deploy your model.",
                                    0.05).start()
        return client, load_object, docker_creds
    except requests.exceptions.HTTPError:
        raise BadCredError()


def begin_packages(local):
    response_containers = requests.get(get_rio_url() + '/containers/?include_stopped=true')
    dfAllContainers = pd.read_json(response_containers.text, orient='records')
    if not dfAllContainers.empty:
        dfStoppedContainers = dfAllContainers[dfAllContainers['State'] == 'exited']
        if not dfStoppedContainers.empty:
            checkDelete = input("Would you like to start all stopped packages? Enter Y to confirm: ").lower()
            if checkDelete == 'y':
                start(local, package_name=None, all=True)


def end(local):
    if local:
        # Get local docker client setup
        # try:
        #     base_url = get_rio_url()
        # except docker.errors.DockerException:
        #     raise DockerNotRunningError
        # # base_url returns none when rio api isn't running locally
        # if not base_url:
        #     click.echo("RIO was not running. To restart, simply run the 'rio begin' command.")
        #     return

        client = APIClient(**kwargs_from_env())

        myContainers = client.containers(all=True)
        for i in myContainers:
            # look for the rio-api container running locally and remove it.
            if i['Names'][0] == '/rio-api':
                client.remove_container(i, force=True)
                click.echo("RIO has ended. To restart, simply run the 'rio begin -l' command.")
    else:
        raise NoLocalFlagError()

    return


def end_packages(local):
    try:
        base_url = get_rio_url()
    except docker.errors.DockerException:
        raise DockerNotRunningError()
    # base_url returns none when rio api isn't running locally
    if not base_url:
        click.echo("RIO isn't running.")
        begin(local=True, file=None)
        base_url = get_rio_url()

    response_containers = requests.get(base_url + '/containers/?include_stopped=true')
    dfAllContainers = pd.read_json(response_containers.text, orient='records')
    if not dfAllContainers.empty:
        dfRunningContainers = dfAllContainers[dfAllContainers['State'] == 'running']
        if not dfRunningContainers.empty:
            checkDelete = input("Would you like to stop all running packages? Enter Y to confirm: ").lower()
            if checkDelete == 'y':
                stop(local, package_name=None, all=True, quick_deploy=False)


def deploy(local, package_name, folder_path, port, quick_deploy):
    # Check if folder name exists
    # folder_path = folder_path.replace('\\', '/')
    full_path = Path(folder_path).resolve()
    if not full_path.exists():
        raise PackagePathError()
    # Get folder name from specified path
    folder_name = full_path.name

    # Use folder name as package name if not specified
    if not package_name:
        package_name = folder_name

    add_to_lib(package_name, full_path)

    # Check if deployment is local
    if local:
        # if port is specified check if it's not used, else get a new unused port
        port, port_is_good = check_port(port)

        if port_is_good:
            # create base url that for the rio api that is running locally to send requests
            try:
                base_url = get_rio_url()
            except docker.errors.DockerException:
                raise DockerNotRunningError()
            # base_url returns none when rio api isn't running locally
            if not base_url:
                click.echo("Spinning up RIO...")
                begin(local=True, file=None)
                base_url = get_rio_url()

            # Get list of packages deployed
            response_packages = requests.get(base_url + '/packages/')
            dfPackages = pd.read_json(response_packages.text, orient='records')

            # Check if there are packages
            if not dfPackages.empty:

                # Check if package/folder_name exists in list
                if package_name in dfPackages.Name.values:
                    click.echo("Package named '" + package_name + "' found.")
                    if not quick_deploy:
                        check_quick_deploy = input(
                            "Would you like to redeploy the package? Enter Y to confirm: ").lower()
                        if check_quick_deploy == 'y':
                            quick_deploy = True
                    if quick_deploy:
                        # get list of containers and split them by running and stopped
                        response_all_containers = requests.get(base_url + '/containers/?include_stopped=true')
                        dfAllContainers = pd.read_json(response_all_containers.text, orient='records')
                        if not dfAllContainers.empty and package_name in dfAllContainers.Names.values:
                            dfRunningContainers = dfAllContainers[dfAllContainers['State'] == 'running']
                            dfStoppedContainers = dfAllContainers[dfAllContainers['State'] == 'exited']
                        else:
                            delete_package(base_url, package_name)
                            deploy(local, package_name, folder_path, port, quick_deploy)
                            return

                        # loader shows progress with begin title, end title and timer. It has been customized.
                        loader1 = loader.Loader("Redeploying Package:",
                                                "The " + package_name + " package is deployed",
                                                "'" + package_name + "' model API has failed to deploy.",
                                                0.05).start()

                        # Check if container exists/running/stopped
                        if package_name in dfRunningContainers.Names.values:

                            # Stop Container
                            response = requests.get(base_url + '/containers/stop/?package_name=' + package_name)
                            if response.status_code == 200:
                                click.echo(" Stopping the Package.")
                        elif package_name in dfStoppedContainers.Names.values:
                            click.echo(" Package is already stopped.")

                        # delete the package
                        response = requests.get(base_url + '/packages/delete/?package_name=' + package_name)
                        if response.status_code == 200:
                            # re-upload package
                            click.echo(" Preparing Package for redeployment...")
                            target_dir = Path(folder_path).resolve().parent

                            # switch to the package directory
                            os.chdir(target_dir)

                            # Compress package
                            archive_path = shutil.make_archive(folder_name, 'zip', target_dir, folder_name)

                            # Mount package to shared path with local rio-api deployment at var/rio-packages
                            # This is also known as the upload step
                            url = base_url + '/upload/?package_name=' + package_name
                            files = {'package': open(archive_path, 'rb')}
                            response = requests.post(url, files=files)
                            if response.status_code == 200:
                                pass
                            else:
                                raise UploadAPIError()

                            files["package"].close()
                            os.remove(archive_path)

                            # restart the api container with updated package
                            response = requests.get(base_url + '/containers/restart/?package_name=' + package_name)

                            if response.status_code == 200:
                                # get the local url of the redeployed package to show the user and launch.
                                launch_url = get_package_url(package_name)
                            else:
                                click.echo("Unable to reload Package.")
                                return

                            logs = None
                            # check if the url is actually up.
                            while not uri_exists_stream(launch_url):
                                # If it doesn't come up and container fails, check if container has stopped.
                                response_containers = requests.get(base_url + '/containers/?include_stopped=true')
                                dfAllContainers = pd.read_json(response_containers.text, orient='records')
                                if not dfAllContainers.empty:
                                    dfStoppedContainers = dfAllContainers[dfAllContainers['State'] == 'exited']
                                    if not dfStoppedContainers.empty:
                                        # If container has stopped save logs and clean up container garbage.
                                        if package_name in dfStoppedContainers.Names.values:
                                            response_logs = requests.get(
                                                base_url + '/containers/logs/?package_name=' + package_name)
                                            logs = response_logs.text
                                            requests.get(
                                                base_url + '/containers/remove/?package_name=' + package_name)
                                            click.echo("Failed to restart the Package.")
                                            break

                            # once out of while loop, assign work based on url being up or down.
                            if uri_exists_stream(launch_url):
                                loader1.stop()
                                if not quick_deploy:
                                    click.echo("at: " + launch_url)
                                    click.pause("Press any key to launch in browser!")
                                    click.launch(launch_url)
                            else:
                                loader1.failed()
                                if logs:
                                    checkLogs = input("Would you like to save the logs?")
                                    if checkLogs == 'Y':
                                        file_name = os.getcwd() + "/" + package_name + "-" + \
                                                    time.strftime("%Y%m%d-%H%M%S") + "-running-logs.txt"
                                        with open(file_name, "w", encoding="utf-8") as log_file:
                                            log_file.write(logs)
                                            log_file.close()
                                        click.echo("Logs are available at: " + file_name)
                                else:
                                    raise LoggingError()
                        else:
                            raise PackageDeletionError()
                    else:
                        raise DuplicatePackageError()
                else:
                    # new deployment path if package name doesn't match deployed packages
                    new_deployment(base_url, full_path, package_name, port, quick_deploy)
            else:
                # new deployment path if no other packages deployed yet
                new_deployment(base_url, full_path, package_name, port, quick_deploy)
        else:
            raise BadPortError()
    else:
        raise NoLocalFlagError()
    return


def new_deployment(base_url, folder_path, package_name, port, quick_deploy):
    click.echo("Preparing package for deployment...")
    # Get folder name from specified path
    folder_name = folder_path.name

    # Use folder name as package name if not specified
    if not package_name:
        package_name = folder_name

    # create the local url of the redeployed package to show the user and launch.
    launch_url = 'http://localhost:' + str(port)

    # loader shows progress with begin title, end title and timer. It has been customized.
    loader1 = loader.Loader("Deploying package:", "The " + package_name + " package is deployed at: " + launch_url,
                            "'" + package_name + "' package has failed to deploy.", 0.05).start()

    target_dir = folder_path.parent

    os.chdir(target_dir)

    # Compress package
    try:
        # Currently failures still create an empty archive, need to find a way to remove them.
        archive_path = shutil.make_archive(folder_name, 'zip', target_dir, folder_name)
    except FileNotFoundError:
        loader1.failed()
        raise PackagingError()

    # Mount package to shared path with local rio-api deployment at var/rio-packages
    # This is also known as the upload step
    url = base_url + '/upload/?package_name=' + package_name
    files = {'package': open(archive_path, 'rb')}
    response = requests.post(url, files=files)

    # Delete the archive we created
    files["package"].close()
    os.remove(archive_path)

    # Check if the API call failed.
    if response.status_code == 404:
        raise UploadAPIError()

    # run the container for the first time locally with our package data in the base image.
    response_run = requests.get(base_url + '/deploy/?package_name=' + package_name + '&port=' + str(port))
    if response_run.status_code == 200:
        click.echo(" Loading package...")
    elif response_run.status_code == 404:
        raise DeployAPIError()

    # check if the url is actually up.
    while not uri_exists_stream(launch_url):
        # If it doesn't come up and container fails, check if container has stopped.
        response_containers = requests.get(base_url + '/containers/?include_stopped=true')
        dfAllContainers = pd.read_json(response_containers.text, orient='records')
        if not dfAllContainers.empty:
            dfStoppedContainers = dfAllContainers[dfAllContainers['State'] == 'exited']
            if not dfStoppedContainers.empty:
                if package_name in dfStoppedContainers.Names.values:
                    # If container has stopped save logs and clean up container garbage.
                    response_logs = requests.get(base_url + '/containers/logs/?package_name=' + package_name)
                    logs = response_logs.text
                    requests.get(base_url + '/containers/remove/?package_name=' + package_name)
                    requests.get(base_url + '/packages/delete/?package_name=' + package_name)
                    click.echo("Failed to start the package.")
                    break

    # once out of while loop, assign work based on url being up or down.
    if uri_exists_stream(launch_url):
        loader1.stop()
        if not quick_deploy:
            click.pause("Press any key to launch in browser!")
            click.launch(launch_url)
    else:
        loader1.failed()
        checkLogs = input("Would you like to save the logs? Enter Y to confirm: ").lower()
        if checkLogs == 'y':
            file_name = os.getcwd() + "/" + package_name + "-" + time.strftime("%Y%m%d-%H%M%S") + "-running-logs.txt"
            with open(file_name, "w", encoding="utf-8") as log_file:
                log_file.write(logs)
                log_file.close()
            click.echo("Logs are available at: " + file_name)
        else:
            raise LoggingError()
    return


def undeploy(local, package_name, all):
    if local:
        try:
            base_url = get_rio_url()
        except docker.errors.DockerException:
            raise DockerNotRunningError()
        # base_url returns none when rio api isn't running locally
        if not base_url:
            click.echo("Spinning up RIO...")
            begin(local=True, file=None)
            base_url = get_rio_url()

        # if all isn't specified and a package_name is given
        if not all and package_name:
            response_containers = requests.get(base_url + '/containers/?include_stopped=true')
            dfAllContainers = pd.read_json(response_containers.text, orient='records')
            if not dfAllContainers.empty:
                dfRunningContainers = dfAllContainers[dfAllContainers['State'] == 'running']
                dfStoppedContainers = dfAllContainers[dfAllContainers['State'] == 'exited']

                if package_name in dfRunningContainers.Names.values:
                    checkDelete = input("Would you like to stop and delete the package? Enter Y to confirm: ").lower()
                    if checkDelete == 'y':
                        pkg_lib_undeploy(package_name)
                        response_stop = requests.get(base_url + '/containers/stop/?package_name=' + package_name)
                        if response_stop.status_code == 200:
                            delete_both(base_url, package_name)
                            click.echo("'" + package_name + "' package undeployed.")
                        else:
                            response_kill = requests.get(
                                base_url + '/containers/kill/?package_name=' + package_name)
                            if response_kill.status_code == 200:
                                delete_both(base_url, package_name)
                                click.echo("'" + package_name + "' package undeployed.")
                            else:
                                click.echo(
                                    "Unable to stop " + package_name + "package. "
                                                                       "Please retry or report this as a bug.")
                    else:
                        click.echo(
                            "'" + package_name + "' is running. Please stop it (e.g : rio stop " + package_name +
                            ") and proceed to delete.")

                elif package_name in dfStoppedContainers.Names.values:
                    delete_both(base_url, package_name)
                    pkg_lib_undeploy(package_name)
                    click.echo("'" + package_name + "' package undeployed.")
                else:
                    raise PackageExistenceError(package_name)
            else:
                response_packages = requests.get(base_url + '/packages/')
                dfPackages = pd.read_json(response_packages.text, orient='records')
                click.echo("--------------------")
                if dfPackages.empty:
                    click.echo("No Packages deployed.")
                    click.echo("--------------------")
                elif package_name in dfPackages.Name.values:
                    delete_package(base_url, package_name)
                    pkg_lib_undeploy(package_name)
                    click.echo("'" + package_name + "' package undeployed.")

        # all is an optional flag that requests to  delete all packages
        elif all:
            checkUndeploy = input(
                "Are you sure you want to undeploy all packages? (This cannot be undone.) Enter Y to confirm: ").lower()
            if checkUndeploy == 'y':
                response_containers = requests.get(base_url + '/containers/?include_stopped=true')
                dfAllContainers = pd.read_json(response_containers.text, orient='records')
                if not dfAllContainers.empty:
                    response_stop_all = requests.get(base_url + '/containers/stop/all/')

                    if response_stop_all.status_code == 200:
                        pass
                    else:
                        click.echo("Unable to stop all packages.")

                    response_kill_all = requests.get(base_url + '/containers/kill/all/')

                    if response_kill_all.status_code == 200:
                        pass
                    else:
                        click.echo("Unable to kill all packages.")

                    response_packages = requests.get(base_url + '/packages/')
                    dfPackages = pd.read_json(response_packages.text, orient='records')
                    package_list = dfPackages['Name'].tolist()
                    for package_name in package_list:
                        if package_name in dfAllContainers.Names.values:
                            delete_container(base_url, package_name)
                            click.echo("Undeployed '" + package_name + "' package.")

                response_packages_prune = requests.get(base_url + '/packages/prune/')

                if response_packages_prune.status_code == 200:
                    os.remove(pkg_lib_path)
                    click.echo("All packages undeployed.")
                else:
                    click.echo("Unable to delete all packages")
            else:
                click.echo("Undeploy all aborted.")
        else:
            raise MissingPackageInputError()
    else:
        raise NoLocalFlagError()

    return


def stop(local, package_name, all, quick_deploy):
    if local:
        try:
            base_url = get_rio_url()
        except docker.errors.DockerException:
            raise DockerNotRunningError()
        # base_url returns none when rio api isn't running locally
        if not base_url:
            click.echo("Spinning up RIO...")
            begin(local=True, file=None)
            base_url = get_rio_url()

        # get list of deployed packages that are running
        response_containers = requests.get(base_url + '/containers/?include_stopped=false')
        dfContainers = pd.read_json(response_containers.text, orient='records')
        if not dfContainers.empty:
            if all:
                for package in dfContainers.Names.values:
                    stop_package(base_url, package, quick_deploy)
            # check if package exists in list and stop it.
            elif package_name:
                if package_name in dfContainers.Names.values:
                    stop_package(base_url, package_name, quick_deploy)
                else:
                    raise PackageExistenceError(package_name)
            else:
                raise MissingPackageInputError()

        else:
            raise NoRunningPackagesError()
    else:
        raise NoLocalFlagError()

    return


def stop_package(base_url, package_name, quick_deploy):
    response = requests.get(base_url + '/containers/stop/?package_name=' + package_name)
    if response.status_code == 200:
        while not get_package_state(package_name) == 'exited':
            click.echo("Updating Status...")
        if not quick_deploy:
            click.echo("The '" + package_name + "' package has been stopped.")
    elif response.status_code == 404:
        click.echo("Unable to stop the '" + package_name + "' package.")


def start(local, package_name, all):
    if local:
        try:
            base_url = get_rio_url()
        except docker.errors.DockerException:
            raise DockerNotRunningError()
        # base_url returns none when rio api isn't running locally
        if not base_url:
            click.echo("Spinning up RIO...")
            begin(local=True, file=None)
            base_url = get_rio_url()

        # Get all containers
        response_containers = requests.get(base_url + '/containers/?include_stopped=true')
        dfContainers = pd.read_json(response_containers.text, orient='records')

        if not dfContainers.empty:
            if all:
                for package in dfContainers.Names.values:
                    start_package(base_url, package, all)
            # check if package exists in list and stop it.
            elif package_name:
                if package_name in dfContainers.Names.values:
                    start_package(base_url, package_name, all)
                else:
                    raise PackageExistenceError(package_name)
            else:
                raise MissingPackageInputError()
            # check if the package shows up in the list
        else:
            raise NoRunningPackagesError()
    else:
        raise NoLocalFlagError()
    return


def start_package(base_url, package_name, all):
    response = requests.get(base_url + '/containers/restart/?package_name=' + package_name)
    if response.status_code == 200:
        launch_url = get_package_url(package_name)

        # wait till package starts up to confirm
        while not uri_exists_stream(launch_url):
            response_containers = requests.get(base_url + '/containers/?include_stopped=true')
            dfAllContainers = pd.read_json(response_containers.text, orient='records')
            if not dfAllContainers.empty:
                dfStoppedContainers = dfAllContainers[dfAllContainers['State'] == 'exited']

                # check if package container has errored out to stop waiting
                if not dfStoppedContainers.empty and package_name in dfStoppedContainers.Names.values:
                    response_logs = requests.get(
                        base_url + '/containers/logs/?package_name=' + package_name)
                    logs = response_logs.text
                    requests.get(base_url + '/containers/remove/?package_name=' + package_name)
                    requests.get(base_url + '/packages/delete/?package_name=' + package_name)
                    click.echo("Failed to start the package.")
                    break

        # once out of while loop, assign work based on url being up or down.
        if uri_exists_stream(launch_url):
            click.echo("The '" + package_name + "' package has been started at: " + launch_url)
            if not all:
                checkLaunch = input(
                    "Would you like to launch it in the default browser? Enter Y to confirm : ").lower()
                if checkLaunch == 'y':
                    click.launch(launch_url)
        else:
            checkLogs = input("Would you like to save the logs? Enter Y to confirm: ").lower()
            if checkLogs == 'y':
                file_name = os.getcwd() + "/" + package_name + "-" + time.strftime(
                    "%Y%m%d-%H%M%S") + "-running-logs.txt"
                with open(file_name, "w", encoding="utf-8") as log_file:
                    log_file.write(logs)
                    log_file.close()
                click.echo("Logs are available at: " + file_name)
            else:
                raise LoggingError()
    elif response.status_code == 404:
        click.echo("Unable to start the '" + package_name + "' package.")


def list_packages(local):
    if local:
        try:
            base_url = get_rio_url()
        except docker.errors.DockerException:
            raise DockerNotRunningError()

        # base_url returns none when rio api isn't running locally
        if not base_url:
            click.echo("RIO is not running.")
            begin(local=True, file=None)
            base_url = get_rio_url()

        # get list of uploaded packages
        response_packages = requests.get(base_url + '/packages/')
        dfPackages = pd.read_json(response_packages.text, orient='records')

        # get list of deployed packages
        response_containers = requests.get(base_url + '/containers/?include_stopped=true')
        dfContainers = pd.read_json(response_containers.text, orient='records')

        df_local_pkgs = get_pkg_lib()

        click.echo("--------------------")
        # somehow give a list back to the user, but only one consolidated list.
        if dfPackages.empty:
            if dfContainers.empty:
                click.echo("No packages deployed.")
            else:
                if len(dfContainers) == 1:
                    click.echo(str(len(dfContainers)) + ' package:')
                elif len(dfContainers) > 1:
                    click.echo(str(len(dfContainers)) + ' packages:')
                click.echo("--------------------")
                # list cleanup
                if not df_local_pkgs.empty:
                    dfContainers = pd.merge(dfContainers, df_local_pkgs, left_on=['Names'], right_on=['package_name'],
                                            how='inner')
                dfContainers['Ports'] = dfContainers['Ports'] \
                    .apply(
                    lambda x: 'http://localhost:' + str(x).split(":")[1].split("-")[0] if not pd.isnull(x) else '')
                dfContainers.drop(['Status', 'package_name'], axis=1, inplace=True)
                dfContainers = dfContainers.rename(
                    columns={"Names": "Name", "Ports": "url", "Created": "Last Started"})
                dfContainers = dfContainers.reset_index(drop=True)
                dfContainers.index = np.arange(1, len(dfContainers) + 1)
                click.echo(dfContainers)
                click.echo('Note: Showing limited information')
        else:

            if dfContainers.empty:
                if len(dfPackages) == 1:
                    click.echo(str(len(dfPackages)) + ' package:')
                elif len(dfPackages) > 1:
                    click.echo(str(len(dfPackages)) + ' packages:')
                click.echo("--------------------")
                dfPackages.index = np.arange(1, len(dfPackages) + 1)
                click.echo(dfPackages)
                click.echo('Note: Showing limited information')
            else:
                # Ideally, this is the list called on at all times.
                # Each package should map to a running or stopped container.
                if len(dfContainers) == 1:
                    click.echo(str(len(dfContainers)) + ' package:')
                elif len(dfContainers) > 1:
                    click.echo(str(len(dfContainers)) + ' packages:')
                click.echo("--------------------")
                dfResult = pd.merge(dfContainers, dfPackages, left_on=['Names'], right_on=['Name'], how='inner')
                if not df_local_pkgs.empty:
                    dfResult = pd.merge(dfResult, df_local_pkgs, left_on=['Names'], right_on=['package_name'],
                                        how='inner')
                # Create clean url that can be clicked.
                dfResult['Ports'] = dfResult['Ports'] \
                    .apply(lambda x: 'http://localhost:' + str(x)
                           .split(":")[1].split("-")[0] if not pd.isnull(x) else 'None')
                dfResult.drop(['LastCreated', 'Name', 'package_name'], axis=1, inplace=True)
                dfResult = dfResult.rename(columns={"Names": "Name", "Ports": "url", "Created": "Last Started",
                                                    "LastModified": "Last Updated"})
                dfResult = dfResult.reset_index(drop=True)
                # Number the packages properly
                dfResult.index = np.arange(1, len(dfResult) + 1)
                click.echo(dfResult)
        click.echo("--------------------")
    else:
        raise NoLocalFlagError()

    return


def logs(local, package_name, tail):
    if local:
        try:
            base_url = get_rio_url()
        except docker.errors.DockerException:
            raise DockerNotRunningError()
        # base_url returns none when rio api isn't running locally
        if not base_url:
            click.echo("Spinning up RIO...")
            begin(local=True, file=None)
            base_url = get_rio_url()

        # get all containers
        response_containers = requests.get(base_url + '/containers/?include_stopped=true')
        dfContainers = pd.read_json(response_containers.text, orient='records')

        # find the container to get logs from
        if not dfContainers.empty:
            if package_name in dfContainers.Names.values:
                # Check if tail log lines has been specified
                if tail:
                    response_logs = requests.get(
                        base_url + '/containers/logs/?package_name=' + package_name + '&tail=' + str(tail))
                else:
                    response_logs = requests.get(base_url + '/containers/logs/?package_name=' + package_name)

                if response_logs.status_code == 200:
                    logs = response_logs.text
                    # Create the log filename with timestamp
                    file_name = os.path.join(os.path.join(os.path.expanduser("~"), "Documents"),
                                             package_name + "-" + time.strftime(
                                                 "%Y%m%d-%H%M%S") + "-running-logs.txt")
                    with open(file_name, "w", encoding="utf-8") as log_file:
                        log_file.write(logs)
                        log_file.close()
                    click.echo("Logs are available at: " + file_name)
                elif response_logs.status_code == 404:
                    raise LoggingError()

            else:
                raise PackageExistenceError(package_name)
        else:
            raise NoRunningPackagesError()
    else:
        raise NoLocalFlagError()
    return


def add_to_lib(package_name, folder_path):
    if os.path.exists(pkg_lib_path):
        # Loading packages from pkg_lib_path
        df_pkg_lib = pd.read_pickle(pkg_lib_path)
        if package_name in df_pkg_lib.package_name.values:
            # package already exists.
            for index, row in df_pkg_lib.iterrows():
                if row['package_name'] == package_name and row['folder_path'] == folder_path:
                    # folder_path is the same.
                    pass
                elif row['package_name'] == package_name and row['folder_path'] != folder_path:
                    # updating folder_path.
                    df_pkg_lib.loc[index, 'folder_path'] = folder_path
        else:
            df_new_pkg_lib = pd.DataFrame({'package_name': package_name,
                                           'folder_path': folder_path}, index=[0])
            df_pkg_lib = df_pkg_lib.append(df_new_pkg_lib, sort=False)
            df_pkg_lib = df_pkg_lib.reset_index(drop=True)
            df_pkg_lib.to_pickle(pkg_lib_path)
    else:
        df_pkg_lib = pd.DataFrame({'package_name': package_name,
                                   'folder_path': folder_path}, index=[0])
        df_pkg_lib.to_pickle(pkg_lib_path)
        # click.echo("Your package details have been stored at: " + pkg_lib_path)
    return


def get_pkg_lib():
    if os.path.exists(pkg_lib_path):
        # Loading packages from pkg_lib_path
        df_pkg_lib = pd.read_pickle(pkg_lib_path)
    else:
        # No Package libraries exist locally.
        df_pkg_lib = pd.DataFrame()
    return df_pkg_lib


def pkg_lib_undeploy(package_name):
    df_pkg_lib = get_pkg_lib()
    df_pkg_lib = df_pkg_lib.set_index("package_name")
    df_pkg_lib = df_pkg_lib.drop(package_name)
    df_pkg_lib = df_pkg_lib.reset_index()
    df_pkg_lib.to_pickle(pkg_lib_path)


def pkg_lib_begin():
    df_pkg_lib = get_pkg_lib()
    if not df_pkg_lib.empty:
        click.echo(df_pkg_lib)
        checkRestore = input(
            "Would you like to restore all the above previously deployed packages? Enter Y to confirm: ").lower()
        if checkRestore == 'y':
            click.echo("Restoring previously deployed packages")
            for index, row in df_pkg_lib.iterrows():
                delete_container(get_rio_url(), row['package_name'])
                deploy(local=True, package_name=row['package_name'], folder_path=row['folder_path'], port=None,
                       quick_deploy=True)
        else:
            for index, row in df_pkg_lib.iterrows():
                delete_container(get_rio_url(), row['package_name'])
                undeploy(local=True, package_name=None, all=True)

    # stop(local=True, package_name=None, all=True, quick_deploy = True)



from click.testing import CliRunner

from rio.commands.cmd_deploy import cli
from rio.utilities import errors


def test_deploy_no_package():
    """
    Tests if RIO properly gives an error if no package is given to deploy.
    """
    runner = CliRunner()
    result = runner.invoke(cli)
    assert isinstance(result.exception, SystemExit)
    assert result.exit_code == 2
    assert "Missing argument" in result.output


def test_deploy_nonlocal():
    """
    Tests that not passing the local flag tells one to contact ChainOpt support.
    """
    runner = CliRunner()
    result = runner.invoke(cli, [r"samples\myProject"])
    assert isinstance(result.exception, errors.NoLocalFlagError)


def test_deploy_bad_package_path():
    """
    Tests error handling when passing through a bad package name.
    """
    runner = CliRunner()
    result = runner.invoke(cli, ["-l", "asdfljk"])
    assert isinstance(result.exception, errors.PackagePathError)

